/*
   Copyright (c) 2008 Ivan Vasin <nisavid@gmail.com>
   Copyright (c) 2006, 2007 Thomas Braxton <kde.braxton@gmail.com>
   Copyright (c) 1999 Preston Brown <pbrown@kde.org>
   Copyright (C) 1997-1999 Matthias Kalle Dalheimer (kalle@kde.org)

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/

#include "kconfigmysql_p.h"

#include <kglobal.h>
#include <kconfig.h>
#include "kconfigbackend.h"
#include "kconfigdata.h"

#include <qdatetime.h>
#include <qsqlerror.h>
#include <qsqlfield.h>
#include <qsqlquery.h>

// For some reason this won't work for me (missing library in CMakeLists.txt?)
/*extern bool kde_kiosk_exception;*/
// so I use this
bool kde_kiosk_exception = false;

KConfigMysqlBackend::KConfigMysqlBackend(QObject*& /*parent*/, const QList<QVariant>& /*args*/)
 : KConfigBackend()
{
    QSqlDatabase db;
    if (QSqlDatabase::contains("kconfig"))
	db = QSqlDatabase::database("kconfig");
    else
	db = QSqlDatabase::addDatabase("QMYSQL", "kconfig");
    dbModel = new QSqlTableModel(this, db);
    dbModel->setEditStrategy( QSqlTableModel::OnManualSubmit ); // cached model, must call dbModel->submitAll() after editing each table

    QSqlField nameFld("name", QVariant::String);
    QSqlField valueFld("value", QVariant::String);
    nameFld.setLength(255);
    valueFld.setLength(255);
    dbRecProto.append(nameFld);
    dbRecProto.append(valueFld);
}

KConfigMysqlBackend::~KConfigMysqlBackend()
{
}

QString KConfigMysqlBackend::warningProlog(const QString &dbInfo)
{
    return QString("KConfigMysql: In database %1: ").arg(dbInfo);
}

KConfigBackend::ParseInfo
KConfigMysqlBackend::parseConfig(const QByteArray& currentLocale, KEntryMap& entryMap, ParseOptions options)
{
    return parseConfig(currentLocale, entryMap, options, false);
}

KConfigBackend::ParseInfo
KConfigMysqlBackend::parseConfig(const QByteArray& currentLocale, KEntryMap& entryMap, ParseOptions options, bool merging)
{
    if (filePath().isEmpty())
        return ParseOk;

    bool bDefault = options&ParseDefaults;
    bool allowExecutableValues = options&ParseExpansions;

    QByteArray currentGroup("<default>");

    // hide the password in the db info string
    QString dbInfo(filePath());
    int authEnd = dbInfo.indexOf('@');
    QString auth = dbInfo.left(authEnd);
    int sep = auth.indexOf(':');
    if (sep >= 0)
	dbInfo.replace(sep+1, authEnd-sep-1, "[password]");

    QList<QByteArray> immutableGroups;

    bool fileOptionImmutable = false;
    bool groupOptionImmutable = false;
    bool groupSkip = false;

    foreach (QString fullGroupName, dbModel->database().tables())
    {
	groupOptionImmutable = fileOptionImmutable;

	currentGroup.clear();
	int start = 0, end;
	do {
	    end = start;
	    for (;;) {
		if (end == fullGroupName.length()) {
		    break;
		}
		if (fullGroupName.at(end) == '\x1d')
		    break;
		end++;
	    }
	    if (end + 1 == fullGroupName.length() && start + 2 == end &&
		fullGroupName.at(start) == '$' && fullGroupName.at(start + 1) == 'i')
	    {
		if (currentGroup.isEmpty())
		    fileOptionImmutable = !kde_kiosk_exception;
		else
		    groupOptionImmutable = !kde_kiosk_exception;
	    }
	    else {
		if (!currentGroup.isEmpty())
		    currentGroup += '\x1d';
		QByteArray namePart = fullGroupName.mid(start, end - start).toUtf8();
		currentGroup += namePart;
	    }
	} while ((start = end + 1) <= fullGroupName.length() && fullGroupName.at(end) == '\x1d');

	groupSkip = entryMap.getEntryOption(currentGroup, 0, 0, KEntryMap::EntryImmutable);

	if (groupSkip && !bDefault)
	    continue;

	if (groupOptionImmutable)
	    // Do not make the groups immutable until the entries from
	    // this file have been added.
	    immutableGroups.append(currentGroup);

	dbModel->setTable(currentGroup);
	if (dbModel->select())
	{
	    for (int i=0; i<dbModel->rowCount(); i++)
	    {
		QSqlRecord record = dbModel->record(i);
		QByteArray aKey = record.value("name").toByteArray(), aValue = record.value("value").toByteArray();
		kDebug() << "encountered entry" << aKey << aValue;

		KEntryMap::EntryOptions entryOptions = 0;
		if (groupOptionImmutable)
		    entryOptions |= KEntryMap::EntryImmutable;

		QByteArray locale;
		int start;
		while ((start = aKey.lastIndexOf('[')) >= 0) {
		    int end = aKey.indexOf(']', start);
		    if (end < 0) {
			qWarning() << warningProlog(dbInfo)
				<< "Invalid entry (missing ']')";
			goto next_entry;
		    } else if (end > start + 1 && aKey.at(start + 1) == '$') { // found option(s)
			int i = start + 2;
			while (i < end) {
			    switch (aKey.at(i)) {
				case 'i':
				    if (!kde_kiosk_exception)
					entryOptions |= KEntryMap::EntryImmutable;
				    break;
				case 'e':
				    if (allowExecutableValues)
					entryOptions |= KEntryMap::EntryExpansion;
				    break;
				case 'd':
				    entryOptions |= KEntryMap::EntryDeleted;
				    aKey = aKey.left(start);
				    entryMap.setEntry(currentGroup, aKey, QByteArray(), entryOptions);
				    goto next_entry;
				default:
				    break;
			    }
			    i++;
			}
		    } else { // found a locale
			if (!locale.isNull()) {
			    qWarning() << warningProlog(dbInfo)
				    << "Invalid entry (second locale!?)";
			    goto next_entry;
			}

			locale = aKey.mid(start + 1,end - start - 1);
		    }
		    aKey.truncate(start);
		}
		if (!locale.isEmpty()) {
		    if (locale != currentLocale) {
			// backward compatibility. C == en_US
			if (locale.at(0) != 'C' || currentLocale != "en_US") {
			    if (merging)
				entryOptions |= KEntryMap::EntryRawKey;
			    else
				goto next_entry; // skip this entry if we're not merging
			}
		    }
		}

		if (options&ParseGlobal)
		    entryOptions |= KEntryMap::EntryGlobal;
		if (bDefault)
		    entryOptions |= KEntryMap::EntryDefault;
		if (!locale.isNull())
		    entryOptions |= KEntryMap::EntryLocalized;
		if (entryOptions & KEntryMap::EntryRawKey) {
		    QByteArray rawKey;
		    rawKey.reserve(aKey.length() + locale.length() + 2);
		    rawKey.append(aKey);
		    rawKey.append('[').append(locale).append(']');
		    entryMap.setEntry(currentGroup, rawKey, aValue, entryOptions);
		} else {
		    entryMap.setEntry(currentGroup, aKey, aValue, entryOptions);
		}
	    next_entry:
		continue;
	    }
	}
    }

    // now make sure immutable groups are marked immutable
    foreach(const QByteArray& group, immutableGroups) {
	entryMap.setEntry(group, QByteArray(), QByteArray(), KEntryMap::EntryImmutable);
    }

    return fileOptionImmutable ? ParseImmutable : ParseOk;
    return ParseOk;
}

void KConfigMysqlBackend::writeEntries(const QByteArray& locale, const QString& /*dbInfo*/,
				       const KEntryMap& map, const QMultiHash<QByteArray, QByteArray> deleteMap, bool defaultGroup, bool &firstEntry)
{
    QSqlDatabase db = dbModel->database();
    QByteArray currentGroup, groupName, keyName, value;
    bool groupIsImmutable = false;
    const KEntryMapConstIterator end = map.constEnd();

    for (KEntryMapConstIterator it = map.constBegin(); it != end; ++it)
    {
	const KEntryKey& key = it.key();

        // Either process the default group or all others
        if ((key.mGroup != "<default>") == defaultGroup)
            continue; // skip

        // the only thing we care about groups is, is it immutable?
        if (key.mKey.isNull()) {
            groupIsImmutable = it->bImmutable;
            continue; // skip
        }

	// Create a table for the group if one does not already exist
        const KEntry& currentEntry = *it;
	if (!defaultGroup && currentGroup != key.mGroup) {
	    currentGroup = key.mGroup;

            for (int start = 0, end;; start = end + 1) {
                end = currentGroup.indexOf('\x1d', start);
                if (end < 0) {
                    int cgl = currentGroup.length();
                    if (currentGroup.at(start) == '$' && cgl - start <= 10) {
                        for (int i = start + 1; i < cgl; i++) {
                            char c = currentGroup.at(i);
                            if (c < 'a' || c > 'z')
                                goto nope;
                        }
			groupName.append("\\x24");
                        start++;
                    }
                  nope:
		    groupName = currentGroup;
                    if (groupIsImmutable) {
                        groupName.append("[$i]");
                    }
                    break;
                } else {
		    groupName = currentGroup.left(end);
                }
	    }

	    if (!firstEntry && !dbModel->submitAll())
		kDebug() << dbModel->lastError();
	    //kDebug() << "ensuring group exists" << groupName;
	    //db.exec("CREATE TABLE IF NOT EXISTS `" + groupName + "` (name VARCHAR(255) PRIMARY KEY, value TEXT)");
	    db.exec("CREATE TABLE IF NOT EXISTS `" + groupName + "` (name TINYTEXT PRIMARY KEY, value TINYTEXT) ENGINE = MyISAM");
	    dbModel->setTable(groupName);
        }

        firstEntry = false;
        // it is data for a group

	keyName = key.mKey;
	if (!key.bRaw) {
	    if (key.bLocal && locale != "C") { // 'C' locale == untranslated
		keyName.append('[');
		keyName.append(locale); // locale tag
		keyName.append(']');
            }
        }
        if (currentEntry.bDeleted) {
            if (currentEntry.bImmutable)
		keyName.append("[$di]"); // Deleted + immutable
            else
		keyName.append("[$d]"); // Deleted
        } else {
            if (currentEntry.bImmutable || currentEntry.bExpand) {
		keyName.append("[$");
                if (currentEntry.bImmutable)
		    keyName.append('i');
                if (currentEntry.bExpand)
		    keyName.append('e');
		keyName.append(']');
            }

	    value = currentEntry.mValue;

	    //kDebug() << "inserting into group" << groupName << "entry" << keyName << value;
	    QSqlRecord rec(dbRecProto);
	    rec.setValue("name", keyName);
	    rec.setValue("value", value);
	    dbModel->setFilter("name='" + keyName + "'");
	    if (dbModel->select() && dbModel->rowCount())
		dbModel->setRecord(0, rec);
	    else
		dbModel->insertRecord(-1, rec);
        }
    }

    if (!groupName.isEmpty() && !dbModel->submitAll())
	kDebug() << dbModel->lastError();

    // remove deleted entries
    foreach (QByteArray gName, deleteMap.keys())
    {
	dbModel->setTable(gName);
	foreach (QByteArray keyName, deleteMap.values(gName))
	{
	    dbModel->setFilter("name='" + keyName + "'");
	    if (dbModel->select() && dbModel->rowCount()) {
		kDebug() << "deleting from group" << gName << "entry" << keyName;
		dbModel->removeRow(0);
	    }
	}
	dbModel->submitAll();
    }
}

void KConfigMysqlBackend::writeEntries(const QByteArray& locale, const QString& dbInfo,
				       const KEntryMap& map, const QMultiHash<QByteArray, QByteArray> deleteMap)
{
    bool firstEntry = true;

    // write default group
    writeEntries(locale, dbInfo, map, QMultiHash<QByteArray, QByteArray>(), true, firstEntry);

    // write all other groups
    writeEntries(locale, dbInfo, map, deleteMap, false, firstEntry);
}

bool KConfigMysqlBackend::writeConfig(const QByteArray& locale, KEntryMap& entryMap,
				      WriteOptions options, const KComponentData &/*data*/)
{
    Q_ASSERT(!filePath().isEmpty());

    KEntryMap writeMap;
    bool bGlobal = options & WriteGlobal;

    {
        ParseOptions opts = ParseExpansions;
        if (bGlobal)
            opts |= ParseGlobal;
        ParseInfo info = parseConfig(locale, writeMap, opts, true);
	if (info != ParseOk) { // either there was an error or the database connection is closed
	    kDebug() << "error parsing existing configuration";
	    return false;
	}
    }

    QMultiHash<QByteArray, QByteArray> deleteMap; // entries to be deleted

    const KEntryMapIterator end = entryMap.end();
    for (KEntryMapIterator it=entryMap.begin(); it != end; ++it) {
        if (!it.key().mKey.isEmpty() && !it->bDirty) // not dirty, doesn't overwrite entry in writeMap. skips default entries, too.
            continue;

        const KEntryKey& key = it.key();

        // only write entries that have the same "globality" as the file
        if (it->bGlobal == bGlobal) {
	    if (!it->bDeleted) {
                writeMap[key] = *it;
	    } else {
                KEntryKey defaultKey = key;
                defaultKey.bDefault = true;
                if (!entryMap.contains(defaultKey)) {
		    // deleted entries must actually be deleted from the database since it is not truncated
		    deleteMap.insert(key.mGroup, key.mKey);
                    writeMap.remove(key);
                } else {
                    writeMap[key] = *it; // otherwise write an explicitly deleted entry
		}
            }
            it->bDirty = false;
        }
    }

    // now writeMap should contain only entries to be written
    // so write it out to the database

    Q_ASSERT(dbModel->database().isOpen());
    writeEntries(locale, filePath(), writeMap, deleteMap);

    return true;
}

bool KConfigMysqlBackend::isWritable() const
{
    if (!filePath().isEmpty())
        return true;
    return false;
}

QString KConfigMysqlBackend::nonWritableErrorMessage() const
{
    // hide the password in the db info string
    QString dbInfo(filePath());
    int authEnd = dbInfo.indexOf('@');
    QString auth = dbInfo.left(authEnd);
    int sep = auth.indexOf(':');
    if (sep >= 0)
	dbInfo.replace(sep+1, authEnd-sep-1, "[password]");
    return i18n("Database \"%1\" not writable.\n", dbInfo);
}

void KConfigMysqlBackend::createEnclosing()
{ }

void KConfigMysqlBackend::setFilePath(const QString& databaseInfo)
{
    if (databaseInfo.isEmpty())
	return;

    QString dbInfo = databaseInfo.mid(databaseInfo.lastIndexOf('/', databaseInfo.lastIndexOf('/') - 1) + 1);
    QSqlDatabase db = dbModel->database();

    int sep = dbInfo.indexOf('@');
    if (sep < 0)
	return;
    QString auth = dbInfo.left(sep);
    sep = auth.indexOf(':');
    if (sep == 0)
	return;
    QString userName = auth.left(sep);
    db.setUserName(userName);
    if (sep > 0)
	db.setPassword(auth.mid(sep+1));

    QStringList connInfo = dbInfo.mid(auth.size()+1).split('/');
    if (connInfo.count() < 2)
	return;
    db.setDatabaseName(connInfo.value(1));
    QString hostInfo = connInfo.value(0);
    sep = hostInfo.indexOf(':');
    if (sep < 0)
    {
	db.setHostName(hostInfo);
	db.setPort(3306);
    }
    else
    {
	db.setHostName(hostInfo.left(sep));
	db.setPort(hostInfo.mid(sep+1).toInt());
    }

    db.open();
    Q_ASSERT(db.isOpen());

    setLocalFilePath(dbInfo);
    setLastModified(QDateTime::currentDateTime());
    setSize(1);
}

KConfigBase::AccessMode KConfigMysqlBackend::accessMode() const
{
    if (filePath().isEmpty())
        return KConfigBase::NoAccess;

    return KConfigBase::ReadOnly;
}

bool KConfigMysqlBackend::lock(const KComponentData& componentData)
{
    Q_ASSERT(!filePath().isEmpty());

    if (!lockFile) {
        lockFile = new KLockFile(filePath() + QLatin1String(".lock"), componentData);
    }

    if (lockFile->lock() == KLockFile::LockStale) // attempt to break the lock
        lockFile->lock(KLockFile::ForceFlag);
    return lockFile->isLocked();
}

void KConfigMysqlBackend::unlock()
{
    lockFile->unlock();
    lockFile.clear();
}

bool KConfigMysqlBackend::isLocked() const
{
    return lockFile && lockFile->isLocked();
}